==> CRTL-0.0.3-py3-none-any.whl: crtl/dataset_derivers/data_derive_pipeline.py <==
# span_linking_dataset#handle indirect links(remove or leave), handle nulls
# span_recognition_dataset#assign all links to one of small concept set (remove if not a child of one of them), handle nulls
# concept_linking_dataset#add all concept ids as a document attribute
# entity_linking_dataset#handle indirect links(remove or leave), merge_spans(leaves or root),handle nulls
# entity_recognition_dataset#assign all links to one of a small concept set (remove if not a child of one of them), merge_spans(leaves or root),handle nulls


# null handling will be different based on if task implementation is end-to-end or modular.
# e.g. entities can be extracted but not linked.
from abc import ABC
from collections import defaultdict
from typing import Dict, Iterable, List, Literal, Union
from spacy.tokens import Span, Doc

from crtl.lkb.lkb import Concept, LexicalKnowledgeBase
from crtl.s3_file_utils.s3_file_utils import cached_path


class DeriverComponent(ABC):
    def __call__(self, spans: List[Span]) -> List[Span]:
        raise NotImplementedError


class DeriverPipeline:
    def __init__(self, components: List[DeriverComponent]):
        self._components = components
        self._output_container_type = (
            "ENT"
            if SpanMerger in [type(component) for component in self._components]
            else "SPAN"
        )

    def __call__(self, docs: Iterable[Doc]) -> None:
        for doc in docs:
            spans = doc.spans["sc"]
            for component in self._components:
                spans = component(spans)
            self._set_spans_or_ents(doc, spans)

    def _set_spans_or_ents(self, doc, spans) -> None:
        if self._output_container_type == "ENT":
            raise NotImplementedError
        elif self._output_container_type == "SPAN":
            raise NotImplementedError


class SpanMerger(DeriverComponent):
    def __init__(self, span_merging_strategy: Literal["NARROW", "BROAD"]):
        self._span_merging_strategy = span_merging_strategy

    @classmethod
    def _contains(self,span_1,span_2):
        return span_1.start <= span_2.start and span_1.end >= span_2.end
    def _filter(self,span_1,span_2):
        if self._span_merging_strategy == "NARROW":
            return self._contains(span_1,span_2)
        elif self._span_merging_strategy == "BROAD":
            return self._contains(span_2,span_1)
    def __call__(self,spans:List[Span]):
        filtered_spans = []
        for span_1 in spans:
            for span_2 in spans:
                if span_1 != span_2:
                    if self._filter(span_1,span_2):
                        print(span_1.text,span_2.text)
                        break
            else:
                filtered_spans.append(span_1)
        return filtered_spans


class ConceptGeneralizer(DeriverComponent):
    def __init__(self, high_level_kb_ids: List[str], lkb: LexicalKnowledgeBase):
        self._generalize_kb_ids = high_level_kb_ids
        self._low_to_high_map = self._generate_low_to_high_mapping(high_level_kb_ids,lkb)
        print(self._low_to_high_map)
    def __call__(self, spans: List[Span]) -> List[Span]:
        for span in spans:
            if self._low_to_high_map[span.kb_id_]:
                span.kb_id_ = self._low_to_high_map[span.kb_id_][0]
            else:
                span.kb_id_ = "NIL"
        return spans
    def _generate_low_to_high_mapping(self,high_level_kb_ids:List[Union[str,int]],lkb:LexicalKnowledgeBase)->Dict[Union[str,int],Union[str,int]]:
        # Function to collect the nodes in a breadth-first traversal
        low_to_high_map = defaultdict(lambda:[])
        for high_level_concept in [lkb.cui_to_concept[kb_id] for kb_id in high_level_kb_ids]:
            low_to_high_map[high_level_concept.concept_id].append(high_level_concept.concept_id)
            descendants = self._get_descendants(high_level_concept,lkb)
            for descendant in descendants:
                low_to_high_map[descendant.concept_id].append(high_level_concept.concept_id)
        return low_to_high_map

    @classmethod
    def _get_descendants(cls,concept:Concept,lkb:LexicalKnowledgeBase)->List[Concept]:
        descendants = []
        visited = set()
        queue = []
        queue.append(concept)
        visited.add(concept.concept_id)
        while queue:
            concept = queue.pop(0)
            for concept_relation in concept.related_concepts:
                child_concept = lkb.cui_to_concept[concept_relation.concept_id]
                if child_concept not in visited:
                    if concept_relation.relation == "inverse_isa":
                        queue.append(child_concept)
                        descendants.append(child_concept)
                    visited.add(concept)
        return descendants
    




class NullRemover(DeriverComponent):
    def __call__(self, spans: List[Span]) -> List[Span]:
        return [span for span in spans if span.kb_id_ != "NIL"]


class IndirectLinkNullifier(DeriverComponent):
    def __call__(self, spans: List[Span]) -> List[Span]:
        for span in spans:
            if span._.is_broad:
                span.kb_id_ = "NIL"
        return spans


class DatasetDeriver:
    _deriver_pipeline: DeriverPipeline

    def __call__(self, docs: Iterable[Doc]) -> None:
        self._deriver_pipeline(docs)


class LinkedSpanDeriver(DatasetDeriver):
    def __init__(self, remove_nulls: bool, nullify_indirect_links: bool):
        pipeline_components: List[DeriverComponent] = []
        if nullify_indirect_links:
            pipeline_components.append(IndirectLinkNullifier())
        if remove_nulls:
            pipeline_components.append(NullRemover())
        self._deriver_pipeline = DeriverPipeline(pipeline_components)


class RecognizedSpanDeriver(DatasetDeriver):
    def __init__(self, remove_nulls: bool, high_level_kb_ids: list[str]):
        pipeline_components: List[DeriverComponent] = []
        lkb = LexicalKnowledgeBase.from_gzipped_jsonl(
            LexicalKnowledgeBase.from_gzipped_jsonl(
                cached_path(
                    "https://databrickstrenddata.s3.amazonaws.com/crtl/lkb/trendnet.gzip"
                )
            )
        )
        pipeline_components.append(ConceptGeneralizer(high_level_kb_ids, lkb))
        if remove_nulls:
            pipeline_components.append(NullRemover())
        self._deriver_pipeline = DeriverPipeline(pipeline_components)


class LinkedEntityDeriver(DatasetDeriver):
    def __init__(
        self,
        nullify_indirect_links: bool,
        span_merging_strategy: Literal["NARROW", "BROAD"],
        remove_nulls: bool,
    ):  # Fixes ordering
        pipeline_components: List[DeriverComponent] = []
        if nullify_indirect_links:
            pipeline_components.append(IndirectLinkNullifier())
        if remove_nulls:
            pipeline_components.append(NullRemover())
        pipeline_components.append(SpanMerger(span_merging_strategy))


class RecognizedEntityDeriver(DatasetDeriver):
    def __init__(
        self,
        span_merging_strategy: Literal["NARROW", "BROAD"],
        high_level_kb_ids: list[str],
        remove_nulls: bool,
    ):  # Fixes Ordering
        lkb = LexicalKnowledgeBase.from_gzipped_jsonl(
            LexicalKnowledgeBase.from_gzipped_jsonl(
                cached_path(
                    "https://databrickstrenddata.s3.amazonaws.com/crtl/lkb/trendnet.gzip"
                )
            )
        )
        pipeline_components: List[DeriverComponent] = []
        pipeline_components.append(ConceptGeneralizer(high_level_kb_ids, lkb))
        if remove_nulls:
            pipeline_components.append(NullRemover())
        pipeline_components.append(SpanMerger(span_merging_strategy))

class TaggedConceptsDeriver:
    Doc.set_extension("kb_ids", default=[])
    def __init__(self,with_broad_links:bool):
        self._with_broad_links = with_broad_links
    def __call__(self, docs: Iterable[Doc]) -> None:
        for doc in docs:
            for span in doc.spans["sc"]:
                if self._test_for_inclusion(span):
                    doc._.kb_ids.append(span.kb_id_)
    def _test_for_inclusion(self,span):
        return span.kb_id_ != "NIL" and (self._with_broad_links or not span._.is_broad)

==> CRTL-0.1.0-py3-none-any.whl: crtl/dataset_derivers/data_derive_pipeline.py <==
# span_linking_dataset#handle indirect links(remove or leave), handle nulls
# span_recognition_dataset#assign all links to one of small concept set (remove if not a child of one of them), handle nulls
# concept_linking_dataset#add all concept ids as a document attribute
# entity_linking_dataset#handle indirect links(remove or leave), merge_spans(leaves or root),handle nulls
# entity_recognition_dataset#assign all links to one of a small concept set (remove if not a child of one of them), merge_spans(leaves or root),handle nulls


# null handling will be different based on if task implementation is end-to-end or modular.
# e.g. entities can be extracted but not linked.
from abc import ABC
from collections import defaultdict
from typing import Dict, Iterable, List, Literal, Union
from spacy.tokens import Span, Doc, SpanGroup

from crtl.lkb.lkb import Concept, LexicalKnowledgeBase
from crtl.s3_file_utils.s3_file_utils import cached_path

# trendnet: LexicalKnowledgeBase = LexicalKnowledgeBase.from_gzipped_jsonl(
#     cached_path("s3://databrickstrenddata/crtl/lkb/trendnet.gzip")
# )


class DeriverComponent(ABC):
    def __call__(self, spans: List[Span]) -> List[Span]:
        raise NotImplementedError


class DeriverPipeline:
    def __init__(self, components: List[DeriverComponent]):
        self._components = components
        self._output_container_type = (
            "ENT"
            if SpanMerger in [type(component) for component in self._components]
            else "SPAN"
        )
        print(self._components)

    def __call__(self, docs: Iterable[Doc]) -> None:
        for doc in docs:
            spans = [span for span in doc.spans["sc"]]
            for component in self._components:
                spans = component(spans)
            self._set_spans_or_ents(doc, spans)

    def _set_spans_or_ents(self, doc: Doc, spans: List[Span]) -> None:
       
        if self._output_container_type == "ENT":
            # print([span.kb_id_ for span in spans])
            doc.set_ents(spans)
            # print(doc.ents)
        elif self._output_container_type == "SPAN":
            group = SpanGroup(doc, name="sc", spans=spans)
            doc.spans["sc"] = group


class SpanMerger(DeriverComponent):
    def __init__(self, span_merging_strategy: Literal["NARROW", "BROAD"]):
        self._span_merging_strategy = span_merging_strategy

    @classmethod
    def _contains(self, span_1, span_2):
        return span_1.start <= span_2.start and span_1.end >= span_2.end

    def _filter(self, span_1, span_2):
        if self._span_merging_strategy == "NARROW":
            return self._contains(span_1, span_2)
        elif self._span_merging_strategy == "BROAD":
            return self._contains(span_2, span_1)

    def __call__(self, spans: List[Span]):
        filtered_spans = []
        for span_1 in spans:
            for span_2 in spans:
                if span_1 != span_2:
                    if self._filter(span_1, span_2):
                        print(span_1.text, span_2.text)
                        break
            else:
                filtered_spans.append(span_1)
        return filtered_spans


class ConceptGeneralizer(DeriverComponent):
    def __init__(self, high_level_kb_ids: List[str], lkb: LexicalKnowledgeBase):
        self._generalize_kb_ids = high_level_kb_ids
        self._low_to_high_map = self._generate_low_to_high_mapping(
            high_level_kb_ids, lkb
        )
        self._lkb = lkb
        # print(self._low_to_high_map)

    def __call__(self, spans: List[Span]) -> List[Span]:
        for span in spans:
            if self._low_to_high_map[span.kb_id_]:
                span.kb_id_ = self._low_to_high_map[span.kb_id_][0]
                span.label_ = self._lkb.cui_to_concept[span.kb_id_].canonical_name
            else:
                span.kb_id_ = "NIL"
                span.label_ = "NIL"
            print(span.kb_id_)
        print([span.kb_id_ for span in spans])
        return spans

    def _generate_low_to_high_mapping(
        self, high_level_kb_ids: List[Union[str, int]], lkb: LexicalKnowledgeBase
    ) -> Dict[Union[str, int], Union[str, int]]:
        # Function to collect the nodes in a breadth-first traversal
        low_to_high_map = defaultdict(lambda: [])
        for high_level_concept in [
            lkb.cui_to_concept[kb_id] for kb_id in high_level_kb_ids
        ]:
            low_to_high_map[high_level_concept.concept_id].append(
                high_level_concept.concept_id
            )
            descendants = self._get_descendants(high_level_concept, lkb)
            for descendant in descendants:
                low_to_high_map[descendant.concept_id].append(
                    high_level_concept.concept_id
                )
        return low_to_high_map

    @classmethod
    def _get_descendants(
        cls, concept: Concept, lkb: LexicalKnowledgeBase
    ) -> List[Concept]:
        descendants = []
        visited = set()
        queue = []
        queue.append(concept)
        visited.add(concept.concept_id)
        while queue:
            concept = queue.pop(0)
            for concept_relation in concept.related_concepts:
                child_concept = lkb.cui_to_concept[concept_relation.concept_id]
                if child_concept not in visited:
                    if concept_relation.relation == "inverse_isa":
                        queue.append(child_concept)
                        descendants.append(child_concept)
                    visited.add(concept)
        return descendants


class NullRemover(DeriverComponent):
    def __call__(self, spans: List[Span]) -> List[Span]:
        return [span for span in spans if span.kb_id_ != "NIL"]


class IndirectLinkNullifier(DeriverComponent):
    def __call__(self, spans: List[Span]) -> List[Span]:
        for span in spans:
            if span._.is_broad:
                span.kb_id_ = "NIL"
        return spans


class DatasetDeriver:
    _deriver_pipeline: DeriverPipeline

    def __call__(self, docs: Iterable[Doc]) -> None:
        self._deriver_pipeline(docs)


class LinkedSpanDeriver(DatasetDeriver):
    def __init__(self, remove_nulls: bool, nullify_indirect_links: bool):
        pipeline_components: List[DeriverComponent] = []
        if nullify_indirect_links:
            pipeline_components.append(IndirectLinkNullifier())
        if remove_nulls:
            pipeline_components.append(NullRemover())
        self._deriver_pipeline = DeriverPipeline(pipeline_components)


class RecognizedSpanDeriver(DatasetDeriver):
    def __init__(
        self,
        remove_nulls: bool,
        high_level_kb_ids: list[str],
        lkb: LexicalKnowledgeBase,
    ):
        pipeline_components: List[DeriverComponent] = []
        pipeline_components.append(ConceptGeneralizer(high_level_kb_ids, lkb))
        if remove_nulls:
            pipeline_components.append(NullRemover())
        self._deriver_pipeline = DeriverPipeline(pipeline_components)


class LinkedEntityDeriver(DatasetDeriver):
    def __init__(
        self,
        nullify_indirect_links: bool,
        span_merging_strategy: Literal["NARROW", "BROAD"],
        remove_nulls: bool,
    ):  # Fixes ordering
        pipeline_components: List[DeriverComponent] = []
        if nullify_indirect_links:
            pipeline_components.append(IndirectLinkNullifier())
        if remove_nulls:
            pipeline_components.append(NullRemover())
        pipeline_components.append(SpanMerger(span_merging_strategy))
        self._deriver_pipeline = DeriverPipeline(pipeline_components)


class RecognizedEntityDeriver(DatasetDeriver):
    def __init__(
        self,
        span_merging_strategy: Literal["NARROW", "BROAD"],
        high_level_kb_ids: list[str],
        remove_nulls: bool,
        lkb: LexicalKnowledgeBase,
    ):  # Fixes Ordering
        pipeline_components: List[DeriverComponent] = []
        pipeline_components.append(ConceptGeneralizer(high_level_kb_ids, lkb))
        if remove_nulls:
            pipeline_components.append(NullRemover())
        pipeline_components.append(SpanMerger(span_merging_strategy))
        self._deriver_pipeline = DeriverPipeline(pipeline_components)


class TaggedConceptsDeriver:
    Doc.set_extension("kb_ids", default=[])

    def __init__(self, with_broad_links: bool):
        self._with_broad_links = with_broad_links

    def __call__(self, docs: Iterable[Doc]) -> None:
        for doc in docs:
            for span in doc.spans["sc"]:
                if self._test_for_inclusion(span):
                    doc._.kb_ids.append(span.kb_id_)

    def _test_for_inclusion(self, span):
        return span.kb_id_ != "NIL" and (self._with_broad_links or not span._.is_broad)
